"""Log."""

from datetime import datetime
import logging
from functools import cached_property, partial

from beartype import beartype
from beartype.typing import Any, Callable, Dict
from pydantic import BaseModel
from rich.console import Console
from rich.text import Text

_DEF_LEVEL = logging.ERROR


class _Styles:
    """Based on `tail-jsonl`."""

    timestamp: str = 'dim grey'
    message: str = ''

    level_error: str = 'red'
    level_warn: str = 'yellow'
    level_info: str = 'green'
    level_debug: str = 'dim blue'

    key: str = 'green'
    value: str = ''

    @cached_property
    def _level_lookup(self) -> Dict[int, str]:
        return {
            logging.CRITICAL: self.level_error,
            logging.ERROR: self.level_error,
            logging.WARNING: self.level_warn,
            logging.INFO: self.level_info,
            logging.DEBUG: self.level_debug,
        }

_LEVEL_TO_NAME = {
    logging.CRITICAL: "EXCEPTION",
    logging.ERROR: "ERROR",
    logging.WARNING: "WARNING",
    logging.INFO: "INFO",
    logging.DEBUG: "DEBUG",
    logging.NOTSET: "NOTSET",
}
"""Mapping to logging level name.

https://docs.python.org/3.11/library/logging.html#logging-levels

"""

_STYLES = _Styles()


@beartype
def _log(message, *, _log_level: int, _this_level: int, _console: Console, **kwargs) -> None:
    """Default log function."""
    if _this_level >= _log_level:
        text = Text()
        text.append(f"{datetime.now()} ", style=_STYLES.timestamp)
        text.append(_LEVEL_TO_NAME.get(_this_level), style=_STYLES._level_lookup.get(_this_level))
        text.append(f" {message}", style=_STYLES.message)
        for key, value in kwargs.items():
            text.append(f' {key}:', style=_STYLES.key)
            text.append(f' {str(value): <10}', style=_STYLES.value)
        _console.print(text)

        if _this_level == logging.CRITICAL:
            _console.print_exception(show_locals=True)
            # # Or:
            # from rich.traceback import install
            # install(show_locals=True)


class _LogSingleton(BaseModel):
    """Store pointer to log function."""

    log: Callable[[Any], None]


_LOG_SINGLETON = _LogSingleton(log=partial(_log, _log_level=_DEF_LEVEL, _console=Console()))


class _Logger:

    @beartype
    def debug(self, message, **kwargs) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.DEBUG, **kwargs)

    @beartype
    def info(self, message, **kwargs) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.INFO, **kwargs)

    @beartype
    def warning(self, message, **kwargs) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.WARNING, **kwargs)

    @beartype
    def error(self, message, **kwargs) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.ERROR, **kwargs)

    @beartype
    def exception(self, message, **kwargs) -> None:
        _LOG_SINGLETON.log(message, _this_level=logging.CRITICAL, **kwargs)


@beartype
def configure_logger(log_level: int = _DEF_LEVEL) -> None:
    """Configure global logger."""
    _LOG_SINGLETON.log = partial(_log, _log_level=log_level, _console=Console())


@beartype
def get_logger() -> _Logger:
    """Retrieve global logger."""
    return _Logger()



"""Track delegated processes with rich progress meters.

Based on: https://www.deanmontgomery.com/2022/03/24/rich-progress-and-multiprocessing

"""

import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from time import sleep
from beartype.typing import Any, Callable, Dict, List
from rich.progress import Progress, BarColumn, TimeRemainingColumn, TimeElapsedColumn
import math
from beartype import beartype


@beartype
def _chunked(data: List[Any], count: int) -> List[List[Any]]:
    # TODO: See below link for other options for chunking
    #   https://realpython.com/how-to-split-a-python-list-into-chunks/
    size = len(data)
    chunk_size, chunk_rem = size // count, size % count
    chunk_size += int(math.ceil(chunk_rem / size))
    return [
        data[ix:ix + chunk_size] for ix in range(0, size, chunk_size)
    ]


@beartype
def pretty_process(delegated_task: Callable[[int, Dict, List], List], *, data: List[Any], num_workers: int = 3) -> List[Any]:
    futures = []  # keep track of the jobs
    with Progress(
        "[progress.description]{task.description}",
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        TimeRemainingColumn(),
        TimeElapsedColumn(),
        refresh_per_second=1,
    ) as progress:
        # Share state between process and workers
        with multiprocessing.Manager() as manager:
            shared_progress = manager.dict()
            totals = {}
            task_id_all = progress.add_task("[green]All jobs progress:")

            with ProcessPoolExecutor(max_workers=num_workers) as executor:
                for ix, chunk in enumerate(_chunked(data, count=num_cpus)):
                    task_id = progress.add_task(f"task {ix}")
                    shared_progress[task_id] = 0
                    totals[task_id] = len(chunk)
                    futures.append(executor.submit(
                        delegated_task, task_id=task_id, shared_progress=shared_progress, data=chunk,
                    ))

                # Update progress bar from shared state
                remaining = len(futures)
                while remaining:
                    n_done = 0
                    for task_id, latest in shared_progress.items():
                        n_done += latest
                        progress.update(task_id, completed=latest, total=totals[task_id])
                    progress.update(task_id_all, completed=n_done, total=len(data))
                    remaining = len(futures) - sum([future.done() for future in futures])

                # Collect results and catch and errors
                return [future.result() for future in futures]


# Note: can't use beartype here
def _long_task(task_id: int, shared_progress: Dict, data: List[Any]) -> None:
    for val in data:
        sleep(1)
        shared_progress[task_id] += 1
    return val


if __name__ == "__main__":
    # Resolve number of cores or specified maximum
    num_cpus = 4
    try:
        import psutil
        num_cpus = psutil.cpu_count(logical=False)
    except Exception as exc:
        print(exc)

    result = pretty_process(_long_task, data=[*range(50)], num_workers=num_cpus)
    print(result)



"""Extend Invoke for Calcipy."""

import sys
from pathlib import Path

from beartype.typing import Any, Callable
from beartype import beartype
from beartype.typing import List
from invoke import Task, Collection, Config, Context, Program
from functools import wraps
from contextlib import suppress
import logging
from .log import configure_logger
from invoke import task as invoke_task
from pydantic import BaseModel
from .log import get_logger

logger = get_logger()


class GlobalTaskOptions(BaseModel):
    """Global Task Options."""

    file_args: List[Path]
    verbose: int


class _ShoalProgram(Program):
    """Customized version of Invoke's `Program`."""

    def print_help(self) -> None:
        """Extend print_help with shoal-specific global configuration.

        https://github.com/pyinvoke/invoke/blob/0bcee75e4a26ad33b13831719c00340ca12af2f0/invoke/program.py#L657-L667

        """
        super().print_help()
        print('Global Task Options:')  # noqa: T201
        print('')  # noqa: T201
        self.print_columns([
            ('*file_args', 'List of Paths available globally to all tasks'),
            ('verbose', 'Globally configure logger verbosity (-vvv for most verbose)'),
        ])
        print('')  # noqa: T201


@beartype
def start_program(pkg_name: str, pkg_version: str, module) -> None:
    """Run the customized Invoke Program.

    FYI: recommendation is to extend the `core_args` method, but this won't parse positional arguments:
    https://docs.pyinvoke.org/en/stable/concepts/library.html#modifying-core-parser-arguments

    """
    # Manipulate 'sys.argv' to hide arguments that invoke can't parse
    file_argv: List[Path] = []
    verbose_argv: int = 1
    sys_argv: List[str] = []
    last_argv = ''
    for argv in sys.argv:
        if not last_argv.startswith('-') and Path(argv).is_file():
            file_argv.append(Path(argv))
        elif argv in {'-v', '-vv', '-vvv', '--verbose'}:
            verbose_argv = argv.count('v')
        else:
            sys_argv.append(argv)
        last_argv = argv
    sys.argv = sys_argv

    class ShoalConfig(Config):

        gto: GlobalTaskOptions = GlobalTaskOptions(file_args=file_argv, verbose=verbose_argv)

    _ShoalProgram(
        name=pkg_name,
        version=pkg_version,
        namespace=Collection.from_module(module),
        config_class=ShoalConfig,
    ).run()


@beartype
def task(*task_args, **task_kwargs) -> Callable[[Any], Task]:
    """Wrapper to accept arguments for an invoke task."""
    @beartype
    def wrapper(func) -> Task:
        """Wraps the decorated task."""
        @invoke_task(*task_args, **task_kwargs)
        @beartype
        @wraps(func)
        def inner(ctx: Context, *args, **kwargs) -> Task:
            """Configure logging, then run actual task."""
            verbose = 2
            with suppress(AttributeError):
                verbose = ctx.config.gto.verbose
            log_lookup = {3: logging.NOTSET, 2: logging.DEBUG, 1: logging.INFO, 0: logging.WARNING}
            raw_log_level = log_lookup.get(verbose)
            configure_logger(log_level=logging.ERROR if raw_log_level is None else raw_log_level)

            summary = func.__doc__.split('\n')[0]
            logger.info(f'Running {func.__name__}', summary=summary)
            logger.debug('Task arguments', args=args, kwargs=kwargs)

            result = func(ctx, *args, **kwargs)

            logger.debug(f'Completed {func.__name__}', result=result)
            return result
        return inner
    return wrapper
